from dataclasses import dataclass
from pathlib import Path

import pytest

from md_dead_link_check.preprocess import LinkInfo
from md_dead_link_check.preprocess import detect_headers
from md_dead_link_check.preprocess import detect_links
from md_dead_link_check.preprocess import find_all_markdowns
from md_dead_link_check.preprocess import process_header_to_fragment
from md_dead_link_check.preprocess import process_md_file


def test_find_all_markdowns():
    files = [
        "md_dead_link_check/__init__.py",
        "tests/test_test.py",
        "action.yml",
        "CHANGELOG.md",
        "README.md",
    ]
    md_files = find_all_markdowns(files)
    refs = [
        "CHANGELOG.md",
        "README.md",
    ]
    refs = [Path(r) for r in refs]
    assert refs == md_files


@pytest.mark.parametrize(
    "header, fragment",
    (
        ("1", "1"),
        ("Header 1", "header-1"),
        ("C++ ext", "c-ext"),
        ("H_I (H)", "h_i-h"),
        ("A `quotes` f", "a-quotes-f"),
        ("H $ maths $", "h--maths-"),
        ("H [text](link)", "h-text"),
        ("H [![text](link)](link)", "h-"),
        ("🙀 header with icon", "-header-with-icon"),
        ("דוגמא", "דוגמא"),
        ("例子", "例子"),
    ),
)
def test_process_header_to_fragment(header, fragment):
    assert process_header_to_fragment(header) == fragment


def test_process_md_file():
    md_info = process_md_file(Path("tests/test_md_files/a.md"), Path(__file__).parent.parent)

    assert md_info.path == Path("tests/test_md_files/a.md")
    assert md_info.fragments == [
        "html-tag",
        "introduction",
        "code-block",
        "formula",
        "grave",
        "links",
        "header-with-quotes-and-math",
        "badge",
        "badge-1",
        "badge-2",
        "some-link-link2",
        "some-tag-asd-",
        "id",
        "id2",
        "id3",
    ]

    ref_links = [
        LinkInfo(
            link="https://github.com/AlexanderDokuchaev",
            location=Path("tests/test_md_files/a.md"),
            line_num=27,
        ),
        LinkInfo(
            link="https://github.com",
            location=Path("tests/test_md_files/a.md"),
            line_num=27,
        ),
        LinkInfo(
            link="./b.md",
            location=Path("tests/test_md_files/a.md"),
            line_num=28,
        ),
        LinkInfo(
            link="b.md",
            location=Path("tests/test_md_files/a.md"),
            line_num=28,
        ),
        LinkInfo(
            link="./d/a.md",
            location=Path("tests/test_md_files/a.md"),
            line_num=29,
        ),
        LinkInfo(
            link="https://github.com",
            location=Path("tests/test_md_files/a.md"),
            line_num=29,
        ),
        LinkInfo(
            link="/tests/test_md_files/d/a.md",
            location=Path("tests/test_md_files/a.md"),
            line_num=30,
        ),
        LinkInfo(
            link="https://github.com/AlexanderDokuchaev/md-dead-link-check/actions/workflows/github_action.yml/badge.svg?branch=main",
            location=Path("tests/test_md_files/a.md"),
            line_num=36,
        ),
        LinkInfo(
            link="https://github.com/AlexanderDokuchaev/md-dead-link-check/actions/workflows/github_action.yml",
            location=Path("tests/test_md_files/a.md"),
            line_num=36,
        ),
        LinkInfo(
            link="#",
            location=Path("tests/test_md_files/a.md"),
            line_num=40,
        ),
        LinkInfo(
            link="./b.md#",
            location=Path("tests/test_md_files/a.md"),
            line_num=42,
        ),
        LinkInfo(
            link="#badge",
            location=Path("tests/test_md_files/a.md"),
            line_num=49,
        ),
        LinkInfo(
            link="#badge-1",
            location=Path("tests/test_md_files/a.md"),
            line_num=50,
        ),
        LinkInfo(
            link="#badge-2",
            location=Path("tests/test_md_files/a.md"),
            line_num=51,
        ),
        LinkInfo(
            link="b.md",
            location=Path("tests/test_md_files/a.md"),
            line_num=53,
        ),
        LinkInfo(
            link="b.md",
            location=Path("tests/test_md_files/a.md"),
            line_num=53,
        ),
        LinkInfo(
            link="mailto:example@example.example",
            location=Path("tests/test_md_files/a.md"),
            line_num=57,
        ),
        LinkInfo(
            link="mailto:example@example.example",
            location=Path("tests/test_md_files/a.md"),
            line_num=59,
        ),
        LinkInfo(
            link="ftp://example.example/example",
            location=Path("tests/test_md_files/a.md"),
            line_num=61,
        ),
        LinkInfo(
            link="https://github.com",
            location=Path("tests/test_md_files/a.md"),
            line_num=62,
        ),
        LinkInfo(
            link="https://github.com",
            location=Path("tests/test_md_files/a.md"),
            line_num=70,
        ),
        LinkInfo(
            link="https://github.com/",
            location=Path("tests/test_md_files/a.md"),
            line_num=70,
        ),
    ]
    assert md_info.links == ref_links


@dataclass
class HeaderTestCase:
    line: str
    header: str

    def __str__(self):
        return self.header


@pytest.mark.parametrize(
    "param",
    (
        HeaderTestCase("# 1", "1"),
        HeaderTestCase("## Header 1", "header-1"),
        HeaderTestCase("### Header 1", "header-1"),
        HeaderTestCase("#### Header 1", "header-1"),
        HeaderTestCase('<a id="head">', "head"),
        HeaderTestCase("# head<T>", "head"),
        HeaderTestCase("# head\<T\>", "headt"),
        HeaderTestCase("# h http://link", "h-httplink"),
        HeaderTestCase("# h <https://link>", "h-httpslink"),
        HeaderTestCase("# h <http://link.com/(ver)>", "h-httplinkcomver"),
        HeaderTestCase("# h <asd> ", "h-"),
        HeaderTestCase("# h \<asd\>", "h-asd"),
        HeaderTestCase("# h http://link <link> \<t\>   <http://link>", "h-httplink--t---httplink"),
    ),
    ids=str,
)
def test_detect_headers(param: HeaderTestCase):
    fragments = []
    detect_headers(param.line, fragments)
    assert param.header == fragments[0]


def test_same_header():
    fragments = []
    detect_headers("## Header", fragments)
    detect_headers("## Header", fragments)
    detect_headers("## Header", fragments)
    assert fragments == ["header", "header-1", "header-2"]


@pytest.mark.parametrize(
    "line, ref",
    (
        ("https://link", ["https://link"]),
        ("http://link", ["http://link"]),
        ("[1](link)", ["link"]),
        ("![1](link)", ["link"]),
        ("[![1](img)](link)", ["img", "link"]),
        ("link http://link and https://link2", ["http://link", "https://link2"]),
        ("<http://link>", ["http://link"]),
        ("<http://link(br)>", ["http://link(br)"]),
        ("[1](<http://link(br)>)", ["http://link(br)"]),
        ("[![1](<http://link(br)>)](link)", ["http://link(br)", "link"]),
    ),
)
def test_detect_links(line, ref):
    ret = detect_links(line)
    assert ret == ref



from __future__ import annotations

import re
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path

from git import Repo

RE_HEADER = r"^(?:\s*[-+*]\s+|)[#]{1,6}\s*(.*?)\s*[#]*$"
RE_URL = r"(http[s]?://[^>)\]\s\"]+)"
RE_URL_IN_BRACKETS = r"<(http[s]?://[^>\s]+)>"
RE_LINK = r"([!]{0,1})\[([^\]!]*)\]\(([^()\s]+(?:\([^()\s]*\))*)\s*(.*?)\)"
RE_HTML_TAG = r"</?\w+[^>]*>"
RE_HTML_TAG_ID = r"<\w+\s+(?:[^>]*?\s+)?(?:id|name)=([\"'])(.*?)\1"
RE_HTML_TAG_HREF = r"<\w+\s+(?:[^>]*?\s+)?href=([\"'])(.*?)\1"
RE_SUB = r"[$`][^`]+?[$`]"

MD_TAG_DISABLE = "<!-- md-dead-link-check: off -->"
MD_TAG_ENABLE = "<!-- md-dead-link-check: on -->"
PLACEHOLDER = "MD_DEAD_LINK_CHECK_PLACEHOLDER"


@dataclass
class LinkInfo:
    link: str
    location: Path
    line_num: int

    def get_location(self) -> str:
        """Returns location link in format path:line"""
        return f"{self.location}:{self.line_num}"

    def __lt__(self, other: LinkInfo) -> bool:
        return self.location < other.location or (self.location == other.location and self.line_num < other.line_num)


@dataclass
class MarkdownInfo:
    path: Path
    fragments: list[str] = field(default_factory=lambda: [])
    links: list[LinkInfo] = field(default_factory=lambda: [])


def find_all_markdowns(all_files: list[str]) -> list[Path]:
    """Filter markdown files."""
    ret = []
    for file_path in all_files:
        p = Path(file_path)
        if p.suffix == ".md":
            ret.append(p)
    return ret


def process_header_to_fragment(header: str) -> str:
    """Converts a Markdown header to a URL fragment."""
    fragment = header.strip()
    fragment = re.sub(RE_HTML_TAG, "", fragment)
    while True:
        res = re.search(RE_LINK, fragment)
        if not res:
            break
        if res.group(1) == "!":
            # Use # to work
            fragment = fragment.replace(res.group(0), "")
        else:
            fragment = fragment.replace(res.group(0), res.group(2))

    fragment = fragment.lower().replace(" ", "-")

    def filter_header_symbols(c: str) -> bool:
        return c.isalpha() or c.isdigit() or c in ["-", "_"]

    return "".join(filter(filter_header_symbols, fragment))


def detect_headers(line: str, fragments: list[str]) -> None:
    """Detect headers in a line and add to the list."""

    # Keep text in \< \>
    # Example:
    #    '# Header \<T\>' becomes 'header-t'
    line = re.sub(r"\\<(.*?)\\>", r"\1", line)

    # Keep links in < >
    # If it does not start with http, the markdown parser will ignore it
    # Example:
    #    '# Header <T>' becomes 'header-'
    line = re.sub(r"(?<!\\)<(http[s]?://[^>\s]+)(?<!\\)>", r"\1", line)

    res = re.match(RE_HEADER, line)
    if res:
        _fragment = process_header_to_fragment(res.group(1))
        fragment = _fragment
        repeat = 0
        while fragment in fragments:
            repeat += 1
            fragment = f"{_fragment}-{repeat}"
        fragments.append(fragment)

    # Skip $ and ` tags
    line = re.sub(RE_SUB, "", line)

    # Detect id under a tag <a id="introduction"></a>
    matches = re.findall(RE_HTML_TAG_ID, line)
    for _, id in matches:
        fragments.append(id.lower())


def detect_links(line: str) -> list[str]:
    """Detect links and paths in a line and return list of links."""
    ret: list[str] = []

    matches = re.findall(RE_URL_IN_BRACKETS, line)
    for link in matches:
        ret.append(link)
        line = line.replace(f"<{link}>", PLACEHOLDER)

    matches = re.findall(RE_LINK, line)
    for _img_tag, _text, link, _title in matches:
        if link != PLACEHOLDER:
            ret.append(link)
        line = line.replace(link, PLACEHOLDER)

    if matches:
        # For case [![text](img_link)](link)
        sub_line = re.sub(RE_LINK, "link", line)
        matches2 = re.findall(RE_LINK, sub_line)
        for _img_tag, _text, link, _title in matches2:
            if link != PLACEHOLDER:
                ret.append(link)
            line = line.replace(link, "")

    # Detect links under a tag <a href="introduction"></a>
    matches = re.findall(RE_HTML_TAG_HREF, line)
    for _, link in matches:
        ret.append(link)
        line = line.replace(link, "")

    # Detect simple urls without any tags
    matches = re.findall(RE_URL, line)
    for url in matches:
        url = re.sub(r"[,.:]+$", "", url)
        ret.append(url)

    return ret


def process_md_file(path: Path, root_dir: Path) -> MarkdownInfo:
    fragments: list[str] = []
    links: list[LinkInfo] = []
    with (root_dir / path).open(encoding="utf8") as stream:
        in_code_block = ""
        disable_detection_links = False
        for line_num, line in enumerate(stream.readlines(), 1):
            striped_line = line.strip()
            # Skip code blocks that can be start ``` or ````
            res = re.match(r"^(`{3,4})(.+)`{3,4}\s*$", striped_line)
            if res:
                continue
            res = re.match(r"^(`{3,4})", striped_line)
            if res and not in_code_block:
                in_code_block = res.group(1)
                continue
            if striped_line.startswith(in_code_block):
                in_code_block = ""
            if in_code_block:
                continue

            # Detect headers
            detect_headers(line, fragments)

            # Skip $ and ` tags
            line = re.sub(RE_SUB, "", line)

            if MD_TAG_DISABLE in line:
                disable_detection_links = True
                continue

            if MD_TAG_ENABLE in line:
                disable_detection_links = False
                continue

            if disable_detection_links:
                continue

            # Detect links
            links_in_line = detect_links(line)
            links.extend(LinkInfo(link, path, line_num) for link in links_in_line)

    return MarkdownInfo(path=path, fragments=fragments, links=links)


def preprocess_repository(untracked_files: bool) -> tuple[dict[str, MarkdownInfo], Path, list[Path]]:
    repo = Repo(search_parent_directories=True)
    root_dir = Path(repo.working_dir)

    all_files: list[str] = repo.git.ls_files().splitlines()
    if untracked_files:
        all_files += repo.untracked_files

    list_md_files = find_all_markdowns(all_files)
    md_data = {}
    for md_file in list_md_files:
        md_info = process_md_file(md_file, root_dir)
        md_data[md_file.as_posix()] = md_info

    files_in_repo = [Path(x) for x in all_files]
    return md_data, root_dir, files_in_repo



import os
import sys
from pathlib import Path

from md_dead_link_check.link_checker import Status
from md_dead_link_check.link_checker import StatusInfo


class SpecSymbols:
    __slots__ = ["blue", "green", "yellow", "red", "clean", "ok", "fail", "split", "cat_ok", "cat_fail"]

    def __init__(self) -> None:
        self.enable()

    def enable(self) -> None:
        self.blue = "\033[1;94m"
        self.green = "\033[1;92m"
        self.yellow = "\033[1;93m"
        self.red = "\033[1;91m"
        self.clean = "\033[0m"

        if sys.platform.startswith("win"):
            self.split = "-"
            self.ok = ""
            self.fail = ""
            self.cat_fail = ""
            self.cat_ok = ""
        else:
            self.split = "•"
            self.ok = "✅ "
            self.fail = "❌ "
            self.cat_fail = " 🙀"
            self.cat_ok = " 😸"

    def disable_colors(self) -> None:
        for key in self.__slots__:
            if key not in ["split"]:
                setattr(self, key, "")


def summary(status: list[StatusInfo], print_warn: bool, print_all: bool, no_color: bool) -> int:
    """Print summary.
    Returns 0 if not found any error, otherwise 1.
    """
    specs = SpecSymbols()
    if no_color:
        specs.disable_colors()
    err_nums = 0
    count_429 = 0

    for x in status:
        link_msg = (
            f"{specs.blue}File:{specs.clean} {x.link_info.get_location()}"
            f" {specs.split} {specs.blue}Link:{specs.clean} {x.link_info.link}"
        )
        if x.msg is not None and "429: too many request" in x.msg.lower():
            count_429 += 1

        if x.status == Status.ERROR:
            print(f"{link_msg} {specs.split} {specs.red}Error{specs.clean}: {x.msg}")
            err_nums += 1
        elif x.status == Status.WARNING and (print_warn or print_all):
            print(f"{link_msg} {specs.split} {specs.yellow}Warn{specs.clean}: {x.msg}")
        elif print_all:
            print(f"{link_msg} {specs.split} {specs.green}OK{specs.clean}")

    if count_429:
        print(
            f"\n{specs.yellow}WARNING:{specs.clean} "
            f'{count_429} link{"s" if count_429 > 1 else ""} returned "429: Too Many Request" respond code. '
            f"This indicates that one of the servers is being accessed too frequently.\n"
            f"To more information visit "
            "https://github.com/AlexanderDokuchaev/md-dead-link-check/#rate-limiting-and-request-throttling"
        )

    if err_nums:
        cat_repeat = 0 if no_color else max(min(err_nums // 10, 5), 1)
        print(f"{specs.fail}Found {err_nums} dead link{'s' if err_nums > 1 else ''}" + specs.cat_fail * cat_repeat)
        return 1
    print(f"{specs.ok}Not found dead links{specs.cat_ok}")
    return 0


def normalize_files(files: list[str], repo_dir: Path) -> list[str]:
    """Set file names to relative git root directory."""
    cwd = Path(os.getcwd())
    if cwd != repo_dir:
        return [(cwd / f).resolve().relative_to(repo_dir).as_posix() for f in files]
    return files



from __future__ import annotations

import asyncio
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from fnmatch import fnmatch
from pathlib import Path
from urllib.parse import urlsplit

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectorError
from aiohttp.client_exceptions import ClientResponseError

from md_dead_link_check.config import Config
from md_dead_link_check.preprocess import LinkInfo
from md_dead_link_check.preprocess import MarkdownInfo

TIMEOUT_RESPONSE_CODE = 408

MSG_TIMEOUT = "408: Timeout"
MSG_PATH_NOT_FOUND = "Path not found"
MSG_PATH_NOT_ADDED = "Path not added to repository"
MSG_FRAGMENT_NOT_FOUND = "Fragment not found"
MSG_UNKNOWN_ERROR = "Unknown error"
MSG_PARSING_ERROR = "Error parsing link"
IGNORED_PROTOCOLS = ("ftp", "sftp")


class Status(int, Enum):
    OK = 0
    WARNING = 1
    ERROR = 2


@dataclass
class StatusInfo:
    link_info: LinkInfo
    status: Status
    msg: str | None = None

    def __lt__(self, other: StatusInfo) -> bool:
        return self.status < other.status or (self.status == other.status and self.link_info < other.link_info)


@dataclass
class LinkStatus:
    link: str
    status: Status
    msg: str | None = None


@dataclass
class LinkWithDelay:
    link: str
    delay: int


async def process_link(data: LinkWithDelay, session: ClientSession, config: Config) -> LinkStatus:
    """Asynchronously processes a link to check its status and gather information.
    Timeout is not interpolated as error, because timeout often occur due to temporary server issues and
    retrying the request might be more appropriate than treating it as an immediate failure.
    """
    link = data.link
    delay = data.delay

    kwargs = {
        "url": link,
        "allow_redirects": True,
        "timeout": config.timeout,
        "ssl": config.validate_ssl,
    }

    try:
        # Use delay to avoid rate limiting (429: Too Many Requests)
        if delay:
            await asyncio.sleep(delay)

        if any(fnmatch(link, p) for p in config.force_get_requests_for_links):
            response = await session.get(**kwargs)
        else:
            response = await session.head(**kwargs)
            if response.status == 404:
                # Some web sites are not supports head request and return 404 code
                response = await session.get(**kwargs)
        response.raise_for_status()
    except ClientResponseError as e:
        if not config.catch_response_codes or e.status in config.catch_response_codes:
            return LinkStatus(link, Status.ERROR, f"{e.status}: {e.message}")
        return LinkStatus(link, Status.WARNING, f"{e.status}: {e.message}")
    except asyncio.CancelledError as e:
        return LinkStatus(link, Status.ERROR, str(e))
    except ClientConnectorError as e:
        return LinkStatus(link, Status.ERROR, str(e))
    except asyncio.TimeoutError:
        if TIMEOUT_RESPONSE_CODE in config.catch_response_codes:
            return LinkStatus(link, Status.ERROR, MSG_TIMEOUT)
        return LinkStatus(link, Status.WARNING, MSG_TIMEOUT)
    except Exception as e:
        msg = str(e)
        if not msg:
            msg = MSG_UNKNOWN_ERROR
        return LinkStatus(link, Status.ERROR, msg)
    return LinkStatus(link, Status.OK)


async def async_check_links(links: list[LinkWithDelay], config: Config) -> list[LinkStatus]:
    async with ClientSession(trust_env=True) as session:
        return await asyncio.gather(*[process_link(li, session, config) for li in links])


def calculate_delay(counter: int, config: Config) -> int:
    return int(min(counter // config.throttle_groups * config.throttle_delay, config.throttle_max_delay))


def generate_delays_for_one_domain_links(links: list[str], config: Config) -> list[LinkWithDelay]:
    domain_requests_counter: dict[str, int] = defaultdict(int)
    ret: list[LinkWithDelay] = []

    for link in links:
        domain = urlsplit(link).netloc
        delay = calculate_delay(domain_requests_counter[domain], config)
        ret.append(LinkWithDelay(link, delay))
        domain_requests_counter[domain] += 1

    enabled_throttling = {d: num - 1 for d, num in domain_requests_counter.items() if num - 1 > config.throttle_groups}
    if enabled_throttling:
        print("Throttling applied to limit request frequency:")
        for domain, num_request in enabled_throttling.items():
            if num_request > config.throttle_groups:
                max_delay = calculate_delay(num_request, config)
                print(f" - Domain:         {domain}")
                print(f"   Requests count: {num_request}")
                if max_delay == config.throttle_max_delay:
                    print(f"   Maximum delay:  {max_delay} seconds (reached throttle_max_delay)")
                else:
                    print(f"   Maximum delay:  {max_delay} seconds")

    return ret


def check_web_links(md_data: dict[str, MarkdownInfo], config: Config, files: list[str]) -> list[StatusInfo]:
    ret: list[StatusInfo] = []

    web_links: list[LinkInfo] = []
    for md_file in files:
        if md_file not in md_data:
            continue
        md_file_info = md_data[md_file]
        if any(fnmatch(md_file, p) for p in config.exclude_files):
            continue
        for li in md_file_info.links:
            try:
                split_result = urlsplit(li.link)
            except ValueError:
                # Dont add error to avoid duplication error message
                continue
            if split_result.scheme in IGNORED_PROTOCOLS:
                continue
            if any(fnmatch(li.link, p) for p in config.exclude_links):
                continue
            if split_result.netloc:
                web_links.append(li)

    # Check only unique links
    unique_links = list(set(li.link for li in web_links))
    links_with_delay = generate_delays_for_one_domain_links(unique_links, config)
    links_status = asyncio.run(async_check_links(links_with_delay, config))

    links_status_dict = {li.link: li for li in links_status}

    for wl in web_links:
        li_status = links_status_dict[wl.link]
        ret.append(StatusInfo(wl, li_status.status, li_status.msg))
    return ret


def check_path_links(
    md_data: dict[str, MarkdownInfo], root_dir: Path, config: Config, files_in_repo: list[Path]
) -> list[StatusInfo]:
    ret: list[StatusInfo] = []

    for md_file, md_file_info in md_data.items():
        if any(fnmatch(md_file, p) for p in config.exclude_files):
            continue
        md_abs_path = root_dir / md_file_info.path
        for md_link in md_file_info.links:
            if md_link.link == "#":
                # Link on top of file
                continue
            if any(fnmatch(md_link.link, p) for p in config.exclude_links):
                continue

            try:
                split_result = urlsplit(md_link.link)
            except ValueError:
                ret.append(StatusInfo(md_link, Status.ERROR, MSG_PARSING_ERROR))
                continue

            if split_result.scheme or split_result.netloc:
                continue
            fragment = split_result.fragment.lower()

            if not split_result.path:
                if fragment not in md_file_info.fragments:
                    ret.append(StatusInfo(md_link, Status.ERROR, MSG_FRAGMENT_NOT_FOUND))
                    continue
            else:
                try:
                    if split_result.path.startswith("/"):
                        # path from git root dir
                        abs_path = root_dir / split_result.path[1:]
                        rel_path = Path(split_result.path[1:])
                    else:
                        abs_path = (md_abs_path.parent / split_result.path).resolve()
                        rel_path = abs_path.relative_to(root_dir)
                except ValueError:
                    ret.append(StatusInfo(md_link, Status.ERROR, MSG_PATH_NOT_FOUND))
                    continue

                if abs_path.as_posix() != abs_path.resolve().as_posix():
                    ret.append(StatusInfo(md_link, Status.ERROR, MSG_PATH_NOT_FOUND))
                    continue

                if rel_path.as_posix() in md_data:
                    # Markdowns in repository
                    if fragment and fragment not in md_data[rel_path.as_posix()].fragments:
                        ret.append(StatusInfo(md_link, Status.ERROR, MSG_FRAGMENT_NOT_FOUND))
                        continue
                else:
                    # Not markdown file
                    if not any(f.as_posix().startswith(rel_path.as_posix()) for f in files_in_repo):
                        if rel_path.exists():
                            ret.append(StatusInfo(md_link, Status.ERROR, MSG_PATH_NOT_ADDED))
                        else:
                            ret.append(StatusInfo(md_link, Status.ERROR, MSG_PATH_NOT_FOUND))
                        continue

            ret.append(StatusInfo(md_link, Status.OK))
    return ret


def check_all_links(
    md_data: dict[str, MarkdownInfo], config: Config, root_dir: Path, files: list[str], files_in_repo: list[Path]
) -> list[StatusInfo]:
    status_list: list[StatusInfo] = []
    if config.check_web_links:
        status_list.extend(check_web_links(md_data, config, files))
    status_list.extend(check_path_links(md_data, root_dir, config, files_in_repo))
    return sorted(status_list)



from pathlib import Path

import pytest
from aiohttp import ClientResponseError
from aiohttp import RequestInfo
from aiohttp.client_exceptions import NonHttpUrlClientError
from pytest_mock import MockerFixture
from yarl import URL

from md_dead_link_check.config import Config
from md_dead_link_check.link_checker import LinkWithDelay
from md_dead_link_check.link_checker import MarkdownInfo
from md_dead_link_check.link_checker import Status
from md_dead_link_check.link_checker import StatusInfo
from md_dead_link_check.link_checker import check_all_links
from md_dead_link_check.link_checker import check_web_links
from md_dead_link_check.link_checker import generate_delays_for_one_domain_links
from md_dead_link_check.preprocess import LinkInfo
from md_dead_link_check.preprocess import process_md_file

ERROR_404 = [
    "https://github.com/AlexanderDokuchaev/FAILELINK",
    "https://github.com/AlexanderDokuchaev/FAILED",
    "https://example.com/(bracket)",
    "https://not_exist_github.githubcom/",
]
CLIENT_CONNECTION_ERROR = [
    "error://urls/",
]


class MockResponse:
    def __init__(self):
        self.status = 200
        self.reason = "OK"

    def raise_for_status(self):
        pass


@pytest.fixture(autouse=True)
def session_mock(mocker: MockerFixture) -> None:
    async def get_side_effect(url, *args, **kwargs):
        if url in ERROR_404:
            raise ClientResponseError(
                RequestInfo(url=url, method="GET", headers={}), (), status=404, message="Not Found"
            )
        if url in CLIENT_CONNECTION_ERROR:
            raise NonHttpUrlClientError(URL(url))
        return MockResponse()

    mocker.patch("aiohttp.ClientSession.get", side_effect=get_side_effect)
    mocker.patch("aiohttp.ClientSession.head", side_effect=get_side_effect)


@pytest.mark.parametrize(
    "url, msg",
    (
        ("https://github.com/AlexanderDokuchaev", None),
        ("https://github.com/AlexanderDokuchaev/FAILELINK", "404: Not Found"),
    ),
)
def test_check_link(url, msg):
    config = Config()
    data = {"test.md": MarkdownInfo("test.md", links=[LinkInfo(url, Path("test.md"), 0)])}
    [r] = check_web_links(data, config, ["test.md"])
    assert r.msg == msg


TEST_FILES = [Path("tests/test_md_files/fail.md"), Path("tests/test_md_files/a.md")]


def test_fails():
    path = "tests/test_md_files/fail.md"
    root_dir = Path(__file__).parent.parent
    md_data = {path: process_md_file(Path(path), root_dir)}
    ret = check_all_links(md_data, Config(), root_dir, list(md_data.keys()), TEST_FILES)

    # Output message depends on proxy settings
    ret[1].msg = None
    ret[1].status = None
    ret[7].msg = None
    ret[7].status = None
    ref = [
        StatusInfo(
            link_info=LinkInfo(
                link="https://github.com/AlexanderDokuchaev/FAILED",
                location=Path("tests/test_md_files/fail.md"),
                line_num=3,
            ),
            status=Status.ERROR,
            msg="404: Not Found",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="https://not_exist_github.githubcom/",
                location=Path("tests/test_md_files/fail.md"),
                line_num=4,
            ),
            status=None,
            msg=None,
        ),
        StatusInfo(
            link_info=LinkInfo(link="/test/fail.md1", location=Path("tests/test_md_files/fail.md"), line_num=8),
            status=Status.ERROR,
            msg="Path not found",
        ),
        StatusInfo(
            link_info=LinkInfo(link="fail.md1", location=Path("tests/test_md_files/fail.md"), line_num=9),
            status=Status.ERROR,
            msg="Path not found",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="/tests/test_md_files/fail.md#fail",
                location=Path("tests/test_md_files/fail.md"),
                line_num=13,
            ),
            status=Status.ERROR,
            msg="Fragment not found",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="not_exist_dir",
                location=Path("tests/test_md_files/fail.md"),
                line_num=15,
            ),
            status=Status.ERROR,
            msg="Path not found",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="error://urls/",
                location=Path("tests/test_md_files/fail.md"),
                line_num=17,
            ),
            status=Status.ERROR,
            msg="error://urls/",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="https://example.com/(bracket)",
                location=Path("tests/test_md_files/fail.md"),
                line_num=19,
            ),
            msg=None,
            status=None,
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="https://example.com]https://example.com",
                location=Path("tests/test_md_files/fail.md"),
                line_num=21,
            ),
            status=Status.ERROR,
            msg="Error parsing link",
        ),
    ]
    assert ret == ref


def test_exclude_files():
    path = "tests/test_md_files/fail.md"
    root_dir = Path(__file__).parent.parent
    md_data = {path: process_md_file(Path(path), root_dir)}
    ret = check_all_links(md_data, Config(exclude_files=[path]), root_dir, [path], TEST_FILES)
    assert ret == []


@pytest.mark.parametrize(
    "exclude_links",
    (
        ["https://github.com/AlexanderDokuchaev/FAILED", "fail.md1", "/test/fail.md1"],
        ["https://github.com/AlexanderDokuchaev/*", "*.md1"],
    ),
    ids=["no_wildcard", "wildcard"],
)
def test_exclude_links(exclude_links):
    path = "tests/test_md_files/fail.md"
    root_dir = Path(__file__).parent.parent
    md_data = {path: process_md_file(Path(path), root_dir)}
    ret = check_all_links(
        md_data,
        Config(exclude_links=exclude_links),
        root_dir,
        list(md_data.keys()),
        TEST_FILES,
    )

    # Output message depends on proxy settings
    ret[0].status = None
    ret[0].msg = None
    ret[4].status = None
    ret[4].msg = None

    ref = [
        StatusInfo(
            link_info=LinkInfo(
                link="https://not_exist_github.githubcom/", location=Path("tests/test_md_files/fail.md"), line_num=4
            ),
            status=None,
            msg=None,
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="/tests/test_md_files/fail.md#fail",
                location=Path("tests/test_md_files/fail.md"),
                line_num=13,
            ),
            status=Status.ERROR,
            msg="Fragment not found",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="not_exist_dir",
                location=Path("tests/test_md_files/fail.md"),
                line_num=15,
            ),
            status=Status.ERROR,
            msg="Path not found",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="error://urls/",
                location=Path("tests/test_md_files/fail.md"),
                line_num=17,
            ),
            status=Status.ERROR,
            msg="error://urls/",
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="https://example.com/(bracket)",
                location=Path("tests/test_md_files/fail.md"),
                line_num=19,
            ),
            status=None,
            msg=None,
        ),
        StatusInfo(
            link_info=LinkInfo(
                link="https://example.com]https://example.com",
                location=Path("tests/test_md_files/fail.md"),
                line_num=21,
            ),
            status=Status.ERROR,
            msg="Error parsing link",
        ),
    ]
    assert ret == ref


def test_generate_delays_for_one_domain_links():
    links = ["https://example.com/1", "https://example.com/2", "https://example.com/3", "https://example2.com/1"]

    config = Config(throttle_groups=2, throttle_delay=10)
    ret = generate_delays_for_one_domain_links(links, config)
    assert ret == [
        LinkWithDelay("https://example.com/1", 0),
        LinkWithDelay("https://example.com/2", 0),
        LinkWithDelay("https://example.com/3", 10),
        LinkWithDelay("https://example2.com/1", 0),
    ]

    config = Config(throttle_groups=1, throttle_delay=100, throttle_max_delay=1000)
    ret = generate_delays_for_one_domain_links(links, config)
    assert ret == [
        LinkWithDelay("https://example.com/1", 0),
        LinkWithDelay("https://example.com/2", 100),
        LinkWithDelay("https://example.com/3", 200),
        LinkWithDelay("https://example2.com/1", 0),
    ]



import sys
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

PROJECT_NAME = "md_dead_link_check"

DEFAULT_CATCH_RESPONSE_CODES = [
    404,  # Not found
    410,  # Gone
    500,  # Internal Server Error (for cannot connect to host under proxy)
]


@dataclass
class Config:
    timeout: int = 5
    catch_response_codes: list[int] = field(default_factory=lambda: DEFAULT_CATCH_RESPONSE_CODES)
    exclude_links: list[str] = field(default_factory=lambda: [])
    exclude_files: list[str] = field(default_factory=lambda: [])
    force_get_requests_for_links: list[str] = field(default_factory=lambda: [])
    check_web_links: bool = True
    validate_ssl: bool = True
    throttle_groups: int = 100
    throttle_delay: int = 20
    throttle_max_delay: int = 100


def get_config(root_dir: Path, config_path: Path | None) -> Config:
    if not config_path:
        config_path = root_dir / "pyproject.toml"
    config = Config()

    if config_path.is_file():
        with open(config_path, "rb") as handle:
            pyproject_toml = tomllib.load(handle)
        toml_config: dict[str, Any] = pyproject_toml.get("tool", {}).get(PROJECT_NAME, {})

        for key, value in toml_config.items():
            if hasattr(config, key):
                setattr(config, key, value)
            else:
                msg = (
                    f"Unexpected config key `{key}` in {config_path.name}. "
                    f"Available keys: [{', '.join(config.__annotations__)}]"
                )
                raise ValueError(msg)
    if not isinstance(config.timeout, int) or config.timeout < 1:
        msg = "`timeout` must be an integer greater than or equal to 1."
        raise ValueError(msg)
    if not isinstance(config.throttle_groups, int) or config.throttle_groups < 1:
        msg = "`throttle_groups` must be an integer greater than or equal to 1."
        raise ValueError(msg)
    if not isinstance(config.throttle_delay, int) or config.throttle_delay < 0:
        msg = "`throttle_delay` must be a non-negative float or integer."
        raise ValueError(msg)
    if not isinstance(config.throttle_max_delay, int) or config.throttle_max_delay < 0:
        msg = "`throttle_max_delay` must be a non-negative integer."
        raise ValueError(msg)
    return config



from argparse import ArgumentParser
from argparse import Namespace
from argparse import RawTextHelpFormatter
from pathlib import Path

from md_dead_link_check.config import get_config
from md_dead_link_check.helpers import normalize_files
from md_dead_link_check.helpers import summary
from md_dead_link_check.link_checker import check_all_links
from md_dead_link_check.preprocess import preprocess_repository


def args_parser() -> Namespace:
    parser = ArgumentParser(
        description="Checks for broken links (dead links) in a Markdown file within a Git repository.",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "files",
        nargs="*",
        help=(
            "List of files to check for web links."
            "\nIf no files are provided (empty list) and the --hook flag is not used, "
            "then all markdown files in the current directory will be checked."
            "\nNote: Internal links within the files will always be checked, regardless of the "
            "provided files, to detect broken links to removed files."
        ),
    )
    parser.add_argument("--config", "-c", type=Path, help="Path to config file.")
    parser.add_argument(
        "--hook",
        action="store_true",
        help=(
            "Run program in pre-commit hook. If not pass list of files, web links will no check, "
            "by default will check in all files."
        ),
    )
    parser.add_argument("--warn", "-w", action="store_true", help="Show warning messages.")
    parser.add_argument("--all", "-a", action="store_true", help="Show all links.")
    parser.add_argument("--no-color", "-nc", action="store_true", help="Disable coloring of output.")
    parser.add_argument("--untrack", action="store_true", help="Check untracked files.")
    return parser.parse_args()


def main() -> int:
    args = args_parser()

    md_data, repo_dir, files_in_repo = preprocess_repository(untracked_files=args.untrack)
    config = get_config(repo_dir, args.config)

    files = normalize_files(args.files, repo_dir)
    if not args.hook and not files:
        files = list(md_data)

    status_list = check_all_links(md_data, config, repo_dir, files, files_in_repo)
    err_num = summary(status_list, args.warn, args.all, args.no_color)

    return int(err_num != 0)


if __name__ == "__main__":
    raise SystemExit(main())
